import re
from collections import OrderedDict
from datetime import timedelta
from typing import List

import petl
import petl as etl
from dateutil.parser import parse as parsedatetime


def parse_datetime(inp, date_fields: List[str]):
    return etl.convert(inp, date_fields, parsedatetime)


def parse_duration(inp):
    return etl.convert(inp, 'dur', lambda v: timedelta(milliseconds=v))


def add_issue_id_from_description(inp):
    def m(row):
        try:
            match = re.search(r'#(\d+)', row['description'])
            return int(match.group(1))
        except AttributeError:
            return None

    return etl.addfield(inp, 'issue_id', m)


def extract_named_objects_to_columns(inp, named_object_columns: List[str]):
    res = inp

    def m(column1: str, prop1: str):
        def mod(row):
            if row.get(column1):
                return row[column1].get(prop1, None)
            return None

        return mod

    for column in named_object_columns:
        for prop in ('id', 'name'):
            res = etl.addfield(res, f'{column}_{prop}', m(column, prop))

    return res


def select_drain_issues(inp, assignee_id: int, drain_cf_id: int):
    def is_drain(fields: list) -> bool:
        return any(map(lambda field: field['id'] == drain_cf_id and field['value'] == '1', fields))

    # custom fields have more selectivity
    inp = etl.select(inp, 'custom_fields', is_drain)
    return etl.selecteq(inp, 'assigned_to_id', assignee_id)


def group_entries_by_day(inp):
    hdr = petl.header(inp)

    agg = OrderedDict()
    for field in hdr:
        # using first found value
        agg[field] = field, next

    agg['dur'] = 'dur', lambda durs: sum(durs, timedelta())
    agg['start'] = 'start', min

    with_day = petl.addfield(inp, 'start_date', lambda row: row.get('start').date())
    index_keys = ('start_date', 'description')
    result = petl.aggregate(with_day, index_keys, agg)
    return petl.cutout(result, 'start_date')



from schema import Schema, Or, Optional

config_schema = Schema({
    'redmine': {
        'token': str,
        'url': str,
    },
    'toggl': {
        'token': str,
        'workspace_id': int,
    },
    Optional('default_attrs'): {
        Optional('sprint_days'): int,
        Optional('rdm_activity_id'): int,
        Optional('rdm_drain_cf_id'): int,
        Optional('group_entries'): bool,
    },
    'project': {
        str: {
            'rdm_project_id': Or(str, int),
            'tgl_project_id': int,
            Optional('sprint_days'): int,
            Optional('rdm_activity_id'): int,
            Optional('rdm_drain_cf_id'): int,
            Optional('group_entries'): bool,
        }
    }
})



import logging

import petl
import typer
from petl import fromdicts

from .utils import setup_http, setup_config
from ..extract import get_toggl_user

tgl = typer.Typer()
logging = logging.getLogger(__name__)


@tgl.callback()
def context(ctx: typer.Context):
    setup_config(ctx, ctx.meta['config_path'])
    setup_http(ctx)


@tgl.command()
def projects(ctx: typer.Context):
    data = fromdicts(get_toggl_user()['projects'])
    data = petl.selecteq(data, 'wid', ctx.meta['config']['toggl']['workspace_id'])
    print(petl.lookall(data))



from datetime import timedelta
from logging import getLogger
from urllib.error import HTTPError
from urllib.request import Request
from urllib.parse import urljoin
from urllib.request import urlopen
from json import dumps

logging = getLogger(__name__)


def to_redmine_time(base_url: str, data, activity_id, user_id, dry=False):
    it = iter(data)
    head = next(it)

    dur = head.index('dur')
    spent = head.index('start')
    issue = head.index('issue_id')
    desc = head.index('description')

    url = urljoin(base_url, '/time_entries.json')

    stats = {
        'loaded': 0,
        'errored': 0
    }

    if dry: logging.info('Using dry run')

    for row in it:
        entry = {
            'user_id': user_id,
            'issue_id': row[issue],
            'spent_on': row[spent].date().isoformat(),
            'hours': row[dur] / timedelta(hours=1),
            'activity_id': activity_id,
            'comments': row[desc]
        }
        logging.info(
            'Entry: [usr:{user_id} issue:{issue_id} spent_on:{spent_on} activity:{activity_id}]'
            ' {hours}h: {comments}'.format(**entry)
        )

        try:
            jsondata = dumps({'time_entry': entry})
            jsondataasbytes = jsondata.encode('utf-8')  # needs to be bytes

            req = Request(url, method='POST')
            req.add_header('Content-Type', 'application/json; charset=utf-8')
            req.add_header('Content-Length', str(len(jsondataasbytes)))

            if not dry:
                logging.debug('Opening connection...')
                urlopen(req, jsondataasbytes)
            else:
                logging.debug('Simulating connection...')

            logging.debug('Entry created!')
            stats['loaded'] += 1
        except HTTPError as e:
            logging.error(
                'Error loading '
                'Entry: [usr:{user_id} issue:{issue_id} spent_on:{spent_on} activity:{activity_id}]'
                ' {hours}h: {comments}'.format(**entry)
            )

            logging.exception(e)
            logging.error(e.fp.read().decode('utf8'))
            stats['errored'] += 1

    return stats



import logging as logging
from pathlib import Path

import typer
from schema import SchemaError
from toml import loads, TomlDecodeError

from ..config import config_schema
from .utils import TapAsserter, setup_http
from ..extract import get_toggl_user, get_redmine_user, get_redmine_project

config_app = typer.Typer()
logging = logging.getLogger(__name__)


@config_app.command()
def validate(ctx: typer.Context):
    tap = TapAsserter()

    path: Path = ctx.meta['config_path']
    tap.check(path.is_file(), f'Using config from {path}')

    config = path.read_text()
    tap.check(len(config) > 0, f'Config is not empty')

    valid = True
    try:
        config = loads(config)
    except TomlDecodeError as e:
        valid = False
        logging.exception(str(e))

    tap.check(valid, f'Config is valid toml')

    valid = True
    try:
        config_schema.validate(config)
    except SchemaError as e:
        valid = False
        logging.exception(str(e))
    tap.check(valid, 'Config is valid against Schema')

    # checking access
    ctx.meta['config'] = config
    setup_http(ctx)

    tgl_user = None

    valid = True
    try:
        tgl_user = get_toggl_user()
    except Exception as e:
        valid = False
        logging.exception(str(e))
    tap.check(valid, 'Has access to toggl user')

    tgl_projects = list(map(lambda proj: proj.get('id'), tgl_user.get('projects', [])))

    valid = True
    try:
        get_redmine_user(config["redmine"]["url"])
    except Exception as e:
        valid = False
        logging.exception(str(e))
    tap.check(valid, 'Has access to redmine user')

    # checking projects
    for project_alias, project in config['project'].items():
        valid = True
        rdm_id = project.get('rdm_project_id')
        try:
            get_redmine_project(config['redmine']['url'], rdm_id)
        except Exception:
            valid = False
        tap.check(valid, f'Redmine project {rdm_id!r} accessable within alias {project_alias!r}')

        tgl_id = project.get("tgl_project_id")
        tap.check(tgl_id in tgl_projects, f'Toggl project {tgl_id} accessable within alias {project_alias!r}')

        # todo check custom redmine fields validness

    tap.stat()

    if not tap.all_valid:
        exit(1)



import urllib.request as urlreq
from logging import getLogger

logging = getLogger(__name__)

__all__ = (
    'password_mgr', 'setup_toggl_auth', 'setup_redmine_auth', 'install', 'USER_AGENT', 'BASE_TOGGL_URL'
)

USER_AGENT = 'tgl2rdm <bano.notit@gmail.com>'
BASE_TOGGL_URL = 'https://api.track.toggl.com'

password_mgr = urlreq.HTTPPasswordMgrWithPriorAuth()


def setup_toggl_auth(user: str, passwd: str):
    logging.debug('Toggl api authorisation set up')
    password_mgr.add_password(None, BASE_TOGGL_URL, user, passwd)
    # without this thing there will be 403 status code from Toggl
    password_mgr.update_authenticated(BASE_TOGGL_URL, True)


def setup_redmine_auth(endpoint: str, user: str, passwd: str):
    logging.debug(f'Redmine api authorisation set up for {endpoint}')
    password_mgr.add_password(None, endpoint, user, passwd)
    password_mgr.update_authenticated(endpoint, True)


def install():
    logging.debug('HTTP authorization manager setup')
    auth_handler = urlreq.HTTPBasicAuthHandler(password_mgr)
    opener = urlreq.build_opener(auth_handler)
    urlreq.install_opener(opener)



import typer
from pathlib import Path
import toml

tgl = typer.Typer()
rdm = typer.Typer()
app = typer.Typer()

app.add_typer(rdm, name='rdm')
app.add_typer(rdm, name='tgl')




if __name__ == '__main__':
    app()



import logging
from logging import config as logging_configurator, INFO
from datetime import datetime, timedelta
from pathlib import Path

import petl
import typer
from toml import loads

from .utils import get_default_config, get_proj_attr, setup_config, setup_http, get_default_logger_config
from .. import extract, transform, utils, load
from .config import config_app
from .tgl import tgl

app = typer.Typer()
app.add_typer(config_app, name='config')
app.add_typer(tgl, name='toggl')
log = logging.getLogger(__name__)


@app.callback()
def main(
        ctx: typer.Context,
        config_path: Path = typer.Option(
            get_default_config,
            resolve_path=True,
        ),
        logger_config: Path = typer.Option(
            get_default_logger_config,
            resolve_path=True,
        ),
):
    ctx.meta['config_path'] = config_path

    if logger_config.exists():
        if logger_config.suffix == '.toml':
            conf = loads(logger_config.read_text())
            logging_configurator.dictConfig(conf)
        elif logger_config.suffix == '.ini':
            logging_configurator.fileConfig(str(logger_config))
    else:
        logging.basicConfig(level=INFO)
        log.info('Using default logger config')


@app.command(help='Synchronize issues from toggl to readmine from --since to --until dates including')
def sync(ctx: typer.Context,
         project: str = typer.Argument(..., help='The name for the project, specified in config file'),
         since: datetime = typer.Option(..., formats=['%Y-%m-%d']),
         until: datetime = typer.Option(..., formats=['%Y-%m-%d']),
         dry: bool = typer.Option(False, help='Use log entries instead of uploading them to redmine'),
         drain: bool = typer.Option(False, help='Use drain issues for entries without specified dest')):
    config = setup_config(ctx, ctx.meta['config_path'])
    setup_http(ctx)

    ctx.meta['rdm_user'] = extract.get_redmine_user(config["redmine"]["url"])

    time_entries = get_toggl_enteries(config, project, since, until)

    issues = get_redmine_issues(config, project, since)

    issue_ids = petl.columns(issues)['id']
    entries_to_load, unset_entries = petl.biselect(time_entries, lambda row: row['issue_id'] in issue_ids)

    if drain and petl.nrows(unset_entries):
        log.info('Using drain')

        drained, unset_entries = drained_entries(ctx, issues, unset_entries, project)

        log.info(f'Drained {petl.nrows(drained)} issues')

        entries_to_load = petl.cat(entries_to_load, drained)

    if petl.nrows(unset_entries):
        log.warning(f'There\'re {petl.nrows(unset_entries)} unset entries')

    if get_proj_attr(config, project, 'group_entries'):
        log.info('Using group by day and description')

        entries_to_load = transform.group_entries_by_day(entries_to_load)

    load.to_redmine_time(
        config["redmine"]["url"],
        entries_to_load,
        activity_id=get_proj_attr(config, project, 'rdm_activity_id'),
        user_id=ctx.meta['rdm_user'].get('id'),
        dry=dry
    )


def drained_entries(ctx: typer.Context, issues, entries, project):
    config = ctx.meta['config']
    empty_entries, unset_entries = petl.biselect(entries, lambda row: row['issue_id'] is None)

    drain_issues = list(
        petl.dicts(
            transform.select_drain_issues(
                issues,
                assignee_id=ctx.meta['rdm_user']['id'],
                drain_cf_id=get_proj_attr(config, project, 'rdm_drain_cf_id')
            )
        )
    )

    if not len(drain_issues):
        log.error('No drain issues found')
        return petl.head(unset_entries, 0), entries

    if len(drain_issues) > 1:
        log.warning(f'Found {len(drain_issues)} drain issues. Will use only first one')

    drain_issue = drain_issues[0]
    drained = petl.addfield(petl.cutout(empty_entries, 'issue_id'), 'issue_id', drain_issue['id'])
    return drained, unset_entries


def get_redmine_issues(config, project, since):
    args = utils.make_sprint_issues_query(project=get_proj_attr(config, project, 'rdm_project_id'),
                                          current_date=since.date(),
                                          sprint_live=timedelta(days=get_proj_attr(config, project, 'sprint_days')))
    issues = extract.from_redmine_issues('https://rm.onlystudio.org/', **args)
    named_object_columns = [
        # 'project', 'tracker', 'status', 'priority', 'author',
        'assigned_to',
        # 'fixed_version'
    ]
    issues = transform.extract_named_objects_to_columns(issues, named_object_columns)
    return issues


def get_toggl_enteries(config, project, since, until):
    if project not in config['project']:
        log.error('No such project in config')
        raise typer.Exit(code=1)
    else:
        project_cfg = config['project'][project]
    time_entries = extract.from_toggl_timeenteries(workspace=config['toggl']['workspace_id'],
                                                   projects=project_cfg['tgl_project_id'],
                                                   since=since.date(),
                                                   until=until.date())
    nrows = petl.nrows(time_entries)
    if nrows == 0:
        log.info('No entries found')
        raise typer.Exit()
    time_entries = transform.parse_datetime(time_entries, ['start', 'end', 'updated'])
    time_entries = transform.parse_duration(time_entries)
    time_entries = transform.add_issue_id_from_description(time_entries)
    return time_entries



from datetime import date, timedelta


def make_sprint_issues_query(project: int, current_date: date, sprint_live: timedelta) -> dict:
    return {
        'project_id': project,
        'start_date': f'><{(current_date - sprint_live).isoformat()}|{current_date.isoformat()}',
        'limit': 1_000_000,
        'status_id': 'open'
    }



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tgl2rdm', 'tgl2rdm.cli']

package_data = \
{'': ['*']}

install_requires = \
['petl>=1.7.4,<2.0.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'schema>=0.7.4,<0.8.0',
 'toml>=0.10.2,<0.11.0',
 'typer[all]>=0.3.2,<0.4.0']

entry_points = \
{'console_scripts': ['t2m = tgl2rdm.__main__:app']}

setup_kwargs = {
    'name': 'tgl2rdm',
    'version': '0.1.4',
    'description': 'Toggl Track to Redmine synchronization util',
    'long_description': '# tgl2rdm\n\n[Toggl Track](https://toggl.com/) -> [Redmine](https://www.redmine.org/) synchonization.\n\n## Installation\n```sh\npip install tgl2rdm\n```\n\n## Configuration\nYou can just copy [`config.example.toml`](./config.example.toml) and [`logger.example.toml`](./logger.example.toml) to `~/.config/tgl2rdm/` as `config.toml` and `logger.toml` representivly.\n',
    'author': 'BANO.notIT',
    'author_email': 'bano.notit@yandex.ru',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/BANOnotIT/tgl2rdm',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import urllib
import urllib.parse
import urllib.parse
import urllib.request
from datetime import date
from json import loads
from typing import List
from typing import Union, Optional
from urllib.error import HTTPError
from logging import getLogger

from petl.io import fromdicts

from .http import USER_AGENT, BASE_TOGGL_URL

rdm_log = getLogger(f'{__name__}.redmine')
tgl_log = getLogger(f'{__name__}.toggl')


def from_toggl_timeenteries(workspace: int, projects: Optional[Union[int, List[int]]] = None,
                            since: Optional[date] = None, until: Optional[date] = None):
    if not projects:
        projects = []

    if type(projects) is not list:
        projects = [projects]

    params = {
        'workspace_id': workspace,
        'since': since,
        'until': until,
        'user_agent': USER_AGENT,
        'project_ids': ','.join(map(str, list(projects))),
    }

    params = {k: v for k, v in params.items() if bool(v)}
    params = urllib.parse.urlencode(params)
    try:
        resp = urllib.request.urlopen(f'{BASE_TOGGL_URL}/reports/api/v2/details?{params}')
    except HTTPError as e:
        tgl_log.exception(e)
        tgl_log.error(e.fp.read())
        raise e

    json = loads(resp.read())
    tgl_log.debug(f"Got {len(json.get('data'))} toggl entries")
    return fromdicts(json.get('data'))


def from_redmine_issues(base_url: str, **kwargs):
    params = {k: v for k, v in kwargs.items() if v is not None}
    params = urllib.parse.urlencode(params)

    try:
        url = urllib.parse.urljoin(base_url, '/issues.json')
        resp = urllib.request.urlopen(f'{url}?{params}')
    except HTTPError as e:
        rdm_log.exception(e)
        rdm_log.error(e.fp.read())
        raise e

    json = loads(resp.read())
    rdm_log.debug(f"Got {len(json.get('issues'))} redmine issues")
    return fromdicts(json.get('issues'))


def get_toggl_user():
    try:
        resp = urllib.request.urlopen(f'{BASE_TOGGL_URL}/api/v8/me?with_related_data=true')
    except HTTPError as e:
        tgl_log.exception(e)
        tgl_log.error(e.fp.read())
        raise e

    json = loads(resp.read())
    tgl_log.debug(f"Got toggl user data")
    return json.get('data')


def get_redmine_user(base_url: str):
    try:
        url = urllib.parse.urljoin(base_url, '/my/account.json')
        resp = urllib.request.urlopen(url)
    except HTTPError as e:
        rdm_log.exception(e)
        rdm_log.error(e.fp.read())
        raise e

    json = loads(resp.read()).get('user')
    rdm_log.debug(f"Got user #{json['id']}: {json['login']!r}")
    return json


def get_redmine_project(base_url: str, pid: Union[int, str]):
    try:
        url = urllib.parse.urljoin(base_url, f'/projects/{str(pid)}.json')
        resp = urllib.request.urlopen(url)
    except HTTPError as e:
        rdm_log.exception(e)
        rdm_log.error(e.fp.read())
        raise e

    json = loads(resp.read())
    rdm_log.debug(f"Got redmine project {json.get('project').get('name')!r}")
    return json.get('project')



import logging
from pathlib import Path

import typer
from toml import loads

from .. import http

logging = logging.getLogger(__name__)


def get_proj_attr(config, project: str, prop: str):
    if prop in config['project'][project]:
        return config['project'][project][prop]
    elif prop in config.get('default_attrs', {}):
        return config['default_attrs'][prop]
    else:
        raise ValueError(f'No {prop!r} found in {project!r} project or in default_attrs, but it\'s required')


def get_default_config():
    return str(Path(typer.get_app_dir('tgl2rdm')) / 'config.toml')


def get_default_logger_config():
    return str(Path(typer.get_app_dir('tgl2rdm')) / 'logger.toml')


def setup_config(ctx: typer.Context, config_path: Path):
    logging.info(f'Using config from {config_path.absolute()}')
    assert config_path.is_file(), 'asdf'
    config = loads(config_path.read_text())
    ctx.meta['config'] = config

    return config


def setup_http(ctx: typer.Context):
    config = ctx.meta['config']

    http.setup_toggl_auth(config["toggl"]["token"], 'api_token')
    http.setup_redmine_auth(config["redmine"]["url"], config["redmine"]["token"], 'api_token')
    http.install()


class TapAsserter:
    counter = 0
    all_valid = True

    def check(self, cond: bool, text: str):
        self.counter += 1

        self.all_valid &= cond

        if cond:
            print(f'ok {self.counter} - {text}')
        else:
            print(f'not ok {self.counter} - {text}')

    def reset(self):
        self.counter = 0

    def stat(self):
        print(f'1..{self.counter}')
